/* ovt_biased.c --- 1/x transfer function for overmaker.c */
/* 27-mar-2004/FK The negative feedback appears positive. We are subtracting
 *                a negative value. Check that and make it positive if needed.
 */
#include <stdio.h>
#include <stdlib.h>

#include "overmaker.h"
#include "overmakerdefs.h"

static char buf[BUFSZ];

/*
 * Sets up static and global symbols in the overdrive.c file generated
 * by overmaker.c
 */

void hdr_biased () {
  commentln ("Variables for the inverted and biased transfer function");
  sprintf (buf, "static float biasBase = %g;", BIAS);
  codeln (buf);
  commentln ("bias and norm are set in function cfg_biased()");
  codeln ("static float bias;");
  codeln ("static float norm;");

#ifdef ADWS_PRE_DIFF
  commentln ("ovt_biased : One sample memory");
  codeln ("static float adwZ;");
  commentln ("ovt_biased : Positive feedback");
  sprintf (buf, "static float adwFb = %g;", ADWS_FB);
  codeln (buf);
#endif /* ADWS_PRE_DIFF */

#ifdef ADWS_POST_DIFF
  codeln ("static float adwZ1;");
#if 0
  codeln ("static float adwZ2;");
#endif
  sprintf (buf, "static float adwFb2 = %g;", ADWS_FB2);
  codeln (buf);
#endif /* ADWS_POST_DIFF */

#ifdef ADWS_GFB
  codeln ("static float adwGfb = -0.6214;");
  codeln ("static float adwGfZ = 0.0;");
#endif /* ADWS_GFB */

#ifdef SAG_EMULATION
  sprintf (buf, "static float sagZgb = %g;", SAG_ZGB);
  codeln (buf);
#endif /* SAG_EMULATION */

}

void cfg_biased () {
  commentln ("Computes the constants for transfer curve");
  codeln ("void cfg_biased (float new_bias) {");
  pushIndent ();
  codeln ("if (0.0 < new_bias) {");
  pushIndent ();
  codeln ("biasBase = new_bias;");
  commentln ("If power sag emulation is enabled bias is set there.");
  codeln ("bias = biasBase;");
  codeln ("norm = 1.0 - (1.0 / (1.0 + (bias * bias)));");
#if 0
  codeln ("printf (\"\\rBIAS=%10.4fNORM=%10.4f\", bias, norm);");
  codeln ("fflush (stdout);");
#endif
  popIndent ();
  codeln ("}");
  popIndent ();
  codeln ("}");
}

void ctl_biased () {
  codeln ("void ctl_biased (unsigned char uc) {");
  pushIndent ();
  codeln ("float u = (float) uc;");
  sprintf (buf, "float v = %g + ((%g - %g) * ((u * u) / 16129.0));",
	   BIAS_LO,
	   BIAS_HI,
	   BIAS_LO);
  codeln (buf);
  codeln ("cfg_biased (v);");
  popIndent ();
  codeln ("}");

#ifdef ADWS_PRE_DIFF
  vspace (3);
  commentln ("ovt_biased:Sets the positive feedback");
  codeln ("void ctl_biased_fb (unsigned char uc) {");
  pushIndent ();
  sprintf (buf, "adwFb = %g * (((float) uc) / 127.0);", ADWS_FB_MAX);
  codeln (buf);
  codeln ("printf (\"\\rFbk=%10.4f\", adwFb);");
  codeln ("fflush (stdout);");
  popIndent ();
  codeln ("}");
#endif /* ADWS_PRE_DIFF */

#ifdef SAG_EMULATION
  vspace (3);
  commentln ("ovt_biased: Sets sag impact");
  codeln ("void ctl_sagtoBias (unsigned char uc) {");
  pushIndent ();
  sprintf (buf, "sagZgb = %g + ((%g - %g) * (((float) uc) / 127.0));",
	   SAG_ZGB_LO,
	   SAG_ZGB_HI,
	   SAG_ZGB_LO);
  codeln (buf);
  codeln ("printf (\"\\rZGB=%10.4f\", sagZgb);");
  codeln ("fflush (stdout);");
  popIndent ();
  codeln ("}");
#endif /* SAG_EMULATION */

#ifdef ADWS_POST_DIFF
  vspace (3);
  commentln ("ovt_biased: Postdiff feedback control");
  codeln ("void ctl_biased_fb2 (unsigned char uc) {");
  pushIndent ();
  sprintf (buf, "adwFb2 = %g * (((float) uc) / 127.0);", ADWS_FB2_MAX);
  codeln (buf);
  codeln ("printf (\"\\rFb2=%10.4f\", adwFb2);");
  codeln ("fflush (stdout);");
  popIndent ();
  codeln ("}");
#endif /* ADWS_POST_DIFF */

#ifdef ADWS_GFB
  vspace (3);
  commentln ("ovt_biased: Global feedback control");
  codeln ("void ctl_biased_gfb (unsigned char uc) {");
  pushIndent ();
  sprintf (buf, "adwGfb = %g * (((float) uc) / 127.0);", ADWS_GFB_MAX);
  codeln (buf);
  codeln ("printf (\"\\rGfb=%10.4f\", adwGfb);");
  codeln ("fflush (stdout);");
  popIndent ();
  codeln ("}");
#endif /* ADWS_GFB */

#ifdef ADWS_FAT_CTRL
  vspace (3);
  commentln ("ovt_biased: Fat control");
  codeln ("void ctl_biased_fat (unsigned char uc) {");
  pushIndent ();
  codeln ("float a;");
  codeln ("if (uc < 64) {");
  pushIndent ();
  codeln ("if (uc < 32) {");
  pushIndent ();
  sprintf (buf, "adwFb = %g;", ADWS_FB);
  codeln (buf);
  sprintf (buf, "adwFb2 = %g + ((%g - %g) * (((float) uc) / 31.0));",
	   ADWS_FB2, ADWS_FB, ADWS_FB2);
  codeln (buf);
  popIndent ();
  codeln ("} else {");
  pushIndent ();
  sprintf (buf, "adwFb = %g + ((%g - %g) * (((float) (uc - 32)) / 31.0));",
	   ADWS_FB, ADWS_FB_MAX, ADWS_FB);
  codeln (buf);
  sprintf (buf, "adwFb2 = %g;", ADWS_FB);
  codeln (buf);
  popIndent ();
  codeln ("}");

  popIndent ();
  codeln ("} else {");
  pushIndent ();
  sprintf (buf, "adwFb = %g;", ADWS_FB_MAX);
  codeln (buf);
  sprintf (buf, "adwFb2 = %g + ((%g - %g) * (((float) (uc - 64)) / 63.0));",
	   ADWS_FB, ADWS_FB_MAX, ADWS_FB);
  codeln (buf);
  popIndent ();
  codeln ("}");

  popIndent ();
  codeln ("}");
#endif /* ADWS_FAT_CTRL */

}

/*
 *
 */
void configSection (int i, void * vp) {
  sprintf (buf,
	   "if (getConfigParameter_fr (\"%s\", cfg, &adwFb, %g, %g));",
	   "overdrive.x-fb1",
	   0.0,
	   ADWS_FB_MAX);
  codeln (buf);

  sprintf (buf,
	   "else if (getConfigParameter_fr (\"%s\", cfg, &adwFb2, %g, %g));",
	   "overdrive.x-fb2",
	   0.0,
	   ADWS_FB2_MAX);
  codeln (buf);

  sprintf (buf,
	   "else if (getConfigParameter_f (\"%s\", cfg, &sagFb));",
	   "overdrive.x-sagfb");
  codeln (buf);

  codeln ("else return 0;");

}

void ini_biased () {
  sprintf (buf, "cfg_biased (%g);", BIAS);
  codeln (buf);
#ifdef ADWS_PRE_DIFF
  /* This is redundant if we have initialized the static var correctly */
  sprintf (buf, "adwFb = %g;", ADWS_FB);
  codeln (buf);
#endif /* ADWS_PRE_DIFF */
}

void xfr_biased (char * xs, char * ys) {
  commentln ("Adaptive linear-non-linear transfer function");

#ifdef ADWS_GFB
  commentln ("Global negative feedback");
  sprintf (buf, "%s -= (adwGfb * adwGfZ);", xs);
  codeln (buf);
#endif /* ADWS_GFB */

#ifdef ADWS_PRE_DIFF
  codeln ("{");
  pushIndent ();
  sprintf (buf, "float temp = %s - adwZ;", xs);
  codeln (buf);
  sprintf (buf, "adwZ = %s + (adwZ * adwFb);", xs);
  codeln (buf);
  sprintf (buf, "%s = temp;", xs);
  codeln (buf);
  popIndent ();
  codeln ("}");
#endif /* ADWS_PRE_DIFF */



  sprintf (buf, "if (%s < 0.0) {", xs);
  codeln (buf);
  pushIndent ();

  /* Temp var to hold value to be squared */
  sprintf (buf, "float x2 = %s - bias;", xs);
  codeln (buf);

  sprintf (buf, "%s = (1.0 / (1.0 + (x2 * x2))) - 1.0 + norm;", ys);
  codeln (buf);

  popIndent ();
  codeln ("} else {");
  pushIndent ();

  /* Temp var to hold value to be squared */
  sprintf (buf, "float x2 = bias + %s;", xs);
  codeln (buf);

  sprintf (buf, "%s = 1.0 - norm - (1.0 / (1.0 + (x2 * x2)));", ys);
  codeln (buf);

  popIndent ();
  codeln ("}");


#ifdef ADWS_POST_DIFF
  codeln ("{");
  pushIndent ();

  sprintf (buf, "float temp = %s + (adwFb2 * adwZ1);", ys);
  codeln (buf);
  sprintf (buf, "%s = temp - adwZ1;", ys);
  codeln (buf);
  codeln ("adwZ1 = temp;");

  popIndent ();
  codeln ("}");
#endif /* ADWS_POST_DIFF */
  
#ifdef ADWS_GFB
  commentln ("Global negative feedback");
  sprintf (buf, "adwGfZ = %s;", ys);
  codeln (buf);
#endif /* ADWS_GFB */
}

/*
 *
 */
void bindCallbacks () {
  bindCallback (MOD_CONFIG, configSection, NULL);
}
